#include <windows.h>
#include <process.h>
#include <dsound.h>
#include <mmsystem.h>
#include <mmreg.h>

#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#pragma  comment(lib, "winmm.lib")
#pragma  comment(lib, "user32.lib")
#pragma  comment(lib, "dsound.lib")
#pragma  comment(lib, "dxguid.lib")

#define HRES(x)     { HRESULT hr; if( (hr = (x)) != S_OK ) { printf("ERROR : %08X\n", hr); break; }}
#define RELEASE(x)  { if ((x)) { (x)->Release(); (x) = NULL; } }

HRESULT DSStream(HWND hWnd, void (*cb_fill)(void *buf, int samples), int cb_done())
{
  typedef float t_snd;
  enum
  {
    NotifyNum    = 6,
    WaitMs       = 1000,
    Bits         = (sizeof(t_snd) * 8),
    Channel      = (2),
    Freq         = (44100),
    Align        = ((Channel * Bits) / 8),
    BytePerSec   = (Freq * Align),
    Samples      = 512,
    OscMax       = 10,
    Gyabo        = 8,
  };
  WAVEFORMATEX wfx = {
    WAVE_FORMAT_IEEE_FLOAT, Channel, Freq, BytePerSec, Align, Bits, 0
  };
  DWORD                Units    = wfx.nBlockAlign * Samples;
  DWORD                Flip     = 0;
  BOOL                 bDone    = FALSE;
  HANDLE               hEvent   = NULL;
  LPDIRECTSOUNDNOTIFY  hNotify  = NULL;
  LPDIRECTSOUND8       hDS      = NULL;
  LPDIRECTSOUNDBUFFER  hBuf     = NULL;
  DSBPOSITIONNOTIFY    dspn[NotifyNum];
  DSBUFFERDESC         dsbdesc =
  {
    sizeof(DSBUFFERDESC), DSBCAPS_LOCSOFTWARE | DSBCAPS_CTRLPOSITIONNOTIFY, Units * NotifyNum, 0, &wfx, DS3DALG_DEFAULT
  };
  do {
    HRES(DirectSoundCreate8(NULL, &hDS, NULL));
    HRES(hDS->SetCooperativeLevel(hWnd, DSSCL_EXCLUSIVE));
    HRES(hDS->CreateSoundBuffer(&dsbdesc, &hBuf, NULL));
    HRES(hBuf->QueryInterface(IID_IDirectSoundNotify, (VOID **)&hNotify));
    hEvent = CreateEvent(NULL, FALSE, FALSE, NULL);
    for(int i = 0; i < NotifyNum; i++) {
      dspn[i].dwOffset     = Units * (i + 1);
      dspn[i].hEventNotify = hEvent;
    }
    dspn[NotifyNum - 1].dwOffset -= 1;
    HRES(hNotify->SetNotificationPositions(NotifyNum, dspn));
    hBuf->Play(0, 0, DSBPLAY_LOOPING);
    while(!cb_done()) {
      LPVOID dest = NULL;
      DWORD  len  = 0;
      if(WaitForSingleObject(hEvent, WaitMs) != WAIT_OBJECT_0) break;
      if(hBuf->Lock((Flip % NotifyNum) * Units, Units, &dest, &len, NULL, NULL, 0) == DS_OK) {
        if(dest) cb_fill(dest, len);
        hBuf->Unlock(dest, len, NULL, 0);
      }
      Flip++;
    }
    
  } while(0);
  
  //Term
  if(hBuf)    hBuf->Stop();
  if(hEvent)  CloseHandle(hEvent);
  if(hNotify) hNotify->Release();
  if(hBuf)    hBuf->Release();
  if(hDS)     hDS->Release();
  return S_OK;
}

void DSTerm()
{
  //WaitForSingleObject(hThread, INFINITE);
}

struct DelayBuffer
{
  enum
  {
    Max = 100000,
  };
  float Buf[Max];
  unsigned long   Rate;
  int   Index;
  void Fill()
  {
    for(int i = 0 ; i < Rate; i++)
      Buf[i] = 0;//frand();
  }
  void Init(unsigned long m)
  {
    Rate = m;
  }
  void Update(float a)
  {
    Buf[Index] = a;
    Index = (Index + 1) % Rate;
  }
  float Sample(unsigned long n = 0)
  {
    return Buf[ (Index + n) % Rate];
  }
};



//http://www.ari-web.com/service/soft/reverb-2.htm
struct Reverb
{
  enum
  {
    CombMax = 4,
    AllMax  = 1,
  };
  DelayBuffer comb[CombMax];
  DelayBuffer all[AllMax];

  float Sample(float a, int index = 0, int character = 0, int lpfnum = 4)
  {
    const int tau[][4][4] = 
    {
      //prime
      {
        {2063, 1847, 1523, 1277},
        {3089, 2927, 2801, 2111},
        {5479, 5077, 4987, 4057},
        {9929, 7411, 4951, 5077},
      },
      {
        {2053, 1867, 1531, 1259},
        {3109, 2939, 2803, 2113},
        {5477, 5059, 4993, 4051},
        {9949, 7393, 4957, 5059},
      }
    };

    const float gain[] = 
    {
       -(0.7733),
       -(0.7623),
       -(0.7513),
       -(0.7403),
    };
    float D = a * 0.1;
    float E = 0;
    
    //Comb
    for(int i = 0 ; i < CombMax; i++)
    {
      DelayBuffer *reb = &comb[i];
      reb->Init(tau[character % 2][index % 4][i]);
      float k = 0;
      float c = 0;
      int LerpMax = lpfnum + 1;
      for(int h = 0 ; h < LerpMax; h++)
        k += reb->Sample(h);
      k /= float(LerpMax);
      c = a + k;
      reb->Update(c * gain[i]);
      E += c;
    }
    
    //all pass
    const int alltau[] = 
    {
       353, 127,
       751, 353,
       353, 157,
      1039, 751,
    };
    
    for(int i = 0 ; i < AllMax; i++)
    {
      DelayBuffer *reb = &all[i];
      reb->Init(alltau[i]);
      float g = 0.7;
      float c = E + a + reb->Sample(0) * -g;
      reb->Update(c);
      E = c * g;
    }
    
    D = (D + E) * 0.5;
    return D;
  }
};

struct Voice
{
  int Wave0, Wave1;
  int Count;
  int A, D, S, R;
  double ph, dph, fq, vol, Mult, pan, lpf;

  Reverb rebL;
  Reverb rebR;

  void On(int note, unsigned long adsr) {
    union {
      struct { unsigned char r, s, d, a; };
      unsigned long adsr;
    } Data;
    int Rate = 0x100;
    Data.adsr = adsr;
    A = Data.a * Rate;
    D = Data.d * Rate;
    S = Data.s * Rate;
    R = Data.r * Rate;
    //printf("A = %d, D = %d, S = %d, R = %d\n", A, D, S, R);
    Count = A + D + S + R;
    fq = 6.0 / 44100.0;
    for(int i = 0; i < note; i++)
      fq *= 1.0594630943592952645618252949463;
    vol = 0;
    ph  = 0;
  }

  void Get(float *buf)
  {
    double w0 = 0;
    double w1 = 0;
    
    if(Count > 0)
    {
      //envgen
      double eg = 0;
           if(R         >= Count) eg = -0.5 / float(R + 1);
      else if(R + S     >= Count) eg =  0.0 / float(S + 1);
      else if(R + S + D >= Count) eg = -0.5 / float(D + 1);
      else                        eg =  1.0 / float(A + 1);
      vol += eg;
      Count--;

      //osc
      #define         PI    3.14159265358979323846
      double          lfo = sin(2 * PI * ph * 0.01333);
      double          t0  = ph + dph;
      double          t1  = ph * (1.0 + Mult);
                      w0  = -1 + (2.0 * (t0 - floor(t0)));
                      w1  = -1 + (2.0 * (t1 - floor(t1)));
      if(Wave0 == 1)  w0  = (w0 < 0) ? -1 : 1;
      if(Wave0 == 2)  w0  = sin(w0 * PI);
      if(Wave1 == 1)  w1  = (w1 < 0) ? -1 : 1;
      if(Wave1 == 2)  w1  = sin(w1 * PI);
      w0 += w1;

      //niz
      static int a = 1;
      static int b = 234567;
      static int c = 890123;
      a += b; b += c; c += a;
      float N = float(a >> 16) / 32768.0f;
      
      //arp
      ph += (fq + N * 0.00053) * ((lfo < 0) ? 1 : 0.5);
    }
    
    //lp
    static float b0 = 0;
    w0 = w0 + (b0 + w0) * (1 - 0.01799);
    b0 = w0;
    
    w1 = w0;
    w0 = w0 * vol * (1.0 - pan) * 0.5;
    w1 = w1 * vol * (1.0 + pan) * 0.5;
    
    //reb
    buf[0] = rebL.Sample(w0, 3, 0, 8);
    buf[1] = rebR.Sample(w1, 3, 1, 8);
  }
};


#define VOICE_MAX 10
Voice voice[VOICE_MAX];

void cb_fill(void *dest, int samples)
{
  float *buf = (float *)dest;
  samples = samples / 4 / 2;
  for(int i = 0 ; i < samples; i++)
  {
    float L = 0;
    float R = 0;
    for(int v = 0; v < VOICE_MAX; v++)
    {
      Voice *w = &voice[v];
      float buf[2] = {0, 0};
      w->Get(buf);
      L += buf[0];
      R += buf[1];
    }
    buf[i * 2 + 0] = L * 0.01;
    buf[i * 2 + 1] = R * 0.01;
  }
  
  
  //test
  #define GetKey(x) (GetAsyncKeyState(x) & 0x8000)
  static int base = 0x48;
  
  voice[0].pan = 0.40 * +1.0;
  voice[1].pan = 0.25 * +1.0;
  voice[2].pan = 0.25 * -1.0;
  voice[3].pan = 0.40 * -1.0;
  
  static double phph = 0.1;
  if(GetKey(VK_LEFT )) base--;
  if(GetKey(VK_RIGHT)) base++;
  if(GetKey(VK_UP  ))  phph += 0.01;
  if(GetKey(VK_DOWN))  phph -= 0.01;
  printf("Base = %d, Phase = %f\r", base, phph);
  
  voice[0].dph   = phph;
  voice[1].dph   = phph;
  voice[2].dph   = phph;
  voice[3].dph   = phph;
  
  voice[0].Mult  = 0.005;
  voice[1].Mult  = 1.003;
  voice[2].Mult  = 1.003;
  voice[3].Mult  = 1.003;
  
  voice[0].Wave0 = 0;
  voice[1].Wave0 = 0;
  voice[2].Wave0 = 0;
  voice[3].Wave0 = 0;
  voice[0].Wave1 = 1;
  voice[1].Wave1 = 1;
  voice[2].Wave1 = 1;
  voice[3].Wave1 = 1;
  
  if(GetKey('A')) voice[0].On(-12 + base +  0, 0x204400DD);
  if(GetKey('S')) voice[1].On(      base +  4, 0x204400DD);
  if(GetKey('D')) voice[2].On(      base +  7, 0x204400DD);
  if(GetKey('F')) voice[3].On(      base + 11, 0x204400DD);
}

int cb_done()
{
  return (GetAsyncKeyState(VK_ESCAPE) & 0x8000);
}

int main()
{
  printf("Quit     -> ESACPE\n");
  printf("Chord    -> A, S, D, F\n");
  printf("Note     -> Left, Right\n");
  DSStream(GetConsoleWindow(), cb_fill, cb_done);
}
  
